﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace UnattendedExample
{
    internal class Program
    {
        /// <summary>
        /// Upload bank account applications and capture customer names
        /// </summary>
        /// <param name="args"></param>
        private static void Main(string[] args)
        {
            var filename = "Banking_eng.pdf";
            if (args != null && args.Length > 0)
            {
                filename = args[0];
            }
            Console.WriteLine("recognition of " + filename);
            PrivateDocumentsRecognition(filename);
            Console.WriteLine("press any key to exit...");
            Console.ReadKey();
        }

        /// <summary>
        /// Process multiple images
        /// <param name="filename">Name of file to process</param>
        /// </summary>
        private static void PrivateDocumentsRecognition(string filename)
        {
            // FlexiCaptureWebServiceSoapClient – name of interface implemented in web service
            // FlexiCapture.
            // Create a URL to refer to the third version of the service.
            var service = new FlexiCapture.FlexiCaptureWebServiceSoapClient();

            // ======= BASIC AUTHENTICATION =======
            // service.ClientCredentials.UserName.UserName = "username";
            // service.ClientCredentials.UserName.Password = "password";

            // To secure your batches against unauthorized access, specify the current user
            // First, get the name of the user visible to the system
            var username = service.GetCurrentUserIdentity();
            // Next, find yourself among the FlexiCapture users
            var userId = service.FindUser(username.Name);
            if (userId <= 0) throw new Exception("Current user not found");

            Console.WriteLine("connected to FlexiCapture, user id: " + userId);

            // Open a new session and start the processing
            OpenWorkSession(service, userId, filename);
        }

        /// <summary>
        /// Opening the new session and starting the processing
        /// </summary>
        /// <param name="service">FlexiCapture service</param>
        /// <param name="userId">User ID</param>
        /// <param name="filename">Name of file to be processed</param>
        private static void OpenWorkSession(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int userId, string filename)
        {
            // Session parameters
            const int roleType = 11; //The operator role on the user's station
            const int stationType = 10; //The user’s station

            // Open a new processing session
            var sessionId = service.OpenSession(roleType, stationType);
            if (sessionId <= 0) throw new Exception("Couldn’t open the session");
            try
            {
                // Open the first batch and continue the processing
                OpenBankingProject(service, userId, sessionId, filename);
            }
            finally
            {
                // The session must always be closed after the processing is complete
                service.CloseSession(sessionId);
            }
        }

        /// <summary>
        /// Opening the first found project and continuing the processing
        /// </summary>
        /// <param name="service">FlexiCapture service</param>
        /// <param name="userId">User ID</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="filename">Name of file to be processed</param>
        private static void OpenBankingProject(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int userId, int sessionId, string filename)
        {
            // Get a list of projects (optional if
            // there’s only one project, in which case
            // it is more convenient to use the GUID)
            var projects = service.GetProjects();
            var projectGuid = "";
            if (projects != null && projects.Count > 0)
            {
                foreach (var project in projects)
                {
                    if (project.Name != "UnattendedExample") continue;
                    projectGuid = project.Guid;
                    break;
                }
            }

            if (string.IsNullOrEmpty(projectGuid))
            {
                throw new Exception("No project named UnattendedExample found. Please upload this project to the server in order to use the example.");
            }

            // Open the first project named UnattendedExample
            var projectId = service.OpenProject(sessionId, projectGuid);
            if (projectId <= 0) throw new Exception("Couldn’t open the project");
            try
            {
                // Open a new batch and continue the processing
                ProcessRecognition(service, userId, sessionId, projectId, filename);
            }
            finally
            {
                // Always close the open batch before ending the processing
                service.CloseProject(sessionId, projectId);
            }
        }

        /// <summary>
        /// Processing the images
        /// </summary>
        /// <param name="service">FlexiCapture service</param>
        /// <param name="userId">User ID</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="projectId">Project ID</param>
        /// <param name="filename">Name of file to be processed</param>
        private static void ProcessRecognition(FlexiCapture.
        FlexiCaptureWebServiceSoapClient service, int userId, int sessionId, int projectId, string filename)
        {
            // Create anew batch. Specify a name for the batch
            // and keep the other properties unchanged
            var batch = new FlexiCapture.Batch { Name = "Sample API Batch" };
            var batchId = service.AddNewBatch(sessionId, projectId, batch, userId);
            if (batchId <= 0) throw new Exception("Couldn’t create the batch");
            try
            {
                AddImageToBatch(service, sessionId, projectId, batchId, filename);

                // Start processing the batch
                service.ProcessBatch(sessionId, batchId);
                Console.WriteLine("recognition task created, waiting ");

                // Wait for the processing to complete
                var percentCompleted = 0;
                while (percentCompleted < 100)
                {
                    Console.CursorLeft = 0;
                    Console.Write(percentCompleted + "%");
                    percentCompleted = service.GetBatchPercentCompleted(batchId);
                    System.Threading.Thread.Sleep(500);
                }
                Console.CursorLeft = 0;
                Console.WriteLine("complete...");

                // Get the results
                var documents = service.GetDocuments(sessionId, batchId);
                if (documents == null) return;

                // XML names required for parsing the XML results
                var docs = XName.Get("Documents", "http://www.abbyy.com/FlexiCapture/Schemas/Export/FormData.xsd");
                var banking = XName.Get("_Banking_eng", "http://www.abbyy.com/FlexiCapture/Schemas/Export/Banking_eng.xsd");

                // Iterate all the results and display the on the screen
                // For each document, only three fields are displayed:
                // Addressing, First Name, Last Name
                foreach (var document in documents)
                {
                    if (document.Id == 0) continue;
                    // Get the XML file with the recognized data
                    var attachedFile = service.LoadDocumentResult(sessionId, batchId, document.Id, "Result.xml");
                    if (attachedFile.Bytes == null) continue;
                    // Open the results in XML
                    var xml = XDocument.Load(new MemoryStream(attachedFile.Bytes));
                    // Container for recognized data
                    var docsElement = xml.Element(docs);
                    if (docsElement == null) continue;
                    var result = docsElement.Element(banking);
                    if (result == null) continue;
                    // Get the recognized first and last name and form of address
                    var addressing = result.Element("_Addressing");
                    var surname = result.Element("_Last_Name");
                    var name = result.Element("_First_Name");
                    // Display the data
                    Console.WriteLine("- " +
                        (addressing == null ? "" : addressing.Value) + " " +
                        (name == null ? "" : name.Value) + " " +
                        (surname == null ? "" : surname.Value)
                        );
                }
            }
            finally
            {
                // Once processed, the batch can be deleted
                service.DeleteBatch(sessionId, batchId);
            }
        }

        /// <summary>
        /// Add images into the batch
        /// </summary>
        /// <param name="service">FlexiCapture service</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="projectId">Project ID</param>
        /// <param name="batchId">Batch ID</param>
        /// <param name="filename">Name of file to be processed</param>
        private static void AddImageToBatch(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int sessionId, int projectId, int batchId, string filename)
        {
            // Open the batch in which to add the images
            service.OpenBatch(sessionId, batchId);
            try
            {
                // Simple uploading of files smaller than 256 KB
                if (new FileInfo(filename).Length < 0x40000)
                {
                    // Simple method for uploading smaller files.
                    // The same method can be used for larger files if the network is fast enough.
                    // Note that the data will be sent in base64,
                    // which increases network loads by 33%.
                    service.AddNewImage(sessionId, batchId, new FlexiCapture.File()
                    {
                        Bytes = File.ReadAllBytes(filename),
                        Name = filename
                    });
                }
                else
                {
                    // Uploading larger files in base64 has drawbacks:
                    // network loads increase by 33%, heavy requests may be blocked 
                    // by IIS or firewalls. If errors occur or connection is lost, 
                    // the file must be resent in its entirety.
                    // Uploading files via the file service API is more efficient.
                    var doc = new FlexiCapture.Document { BatchId = batchId };
                    var file = new FlexiCapture.File { Name = filename };
                    var documentId = service.AddNewDocument(sessionId, doc, file, false, 0);
                    UploadFile(service, sessionId, projectId, batchId, documentId, filename).Wait();
                }
            }
            finally
            {
                service.CloseBatch(sessionId, batchId);
            }
        }

        /// <summary>
        /// Uploading file
        /// </summary>
        /// <param name="service">FlexiCapture service</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="projectId">Project ID</param>
        /// <param name="batchId">Batch ID</param>
        /// <param name="documentId">Document ID</param>
        /// <param name="filename">Name of file to be processed</param>
        /// <returns></returns>
        private static async Task UploadFile(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int sessionId, int projectId,
            int batchId, int documentId, string filename)
        {
            const int objectType = 0;
            var crc = new Crc32();
            using (var fs = File.OpenRead(filename))
            {
                // Uploading portions of 1 MB each
                var buffer = new byte[0x100000];

                var readed = fs.Read(buffer, 0, buffer.Length);
                var checksum = crc.Next(buffer, 0, readed);
                if (readed < buffer.Length)
                {
                    // Uploading smaller files in their entirety
                    var resp = await FileRequest(service, "Save", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename, 0,
                        new ByteArrayContent(buffer, 0, readed));
                    if (resp.StatusCode != HttpStatusCode.OK) throw new Exception("Server error");
                }
                else
                {
                    var offset = 0;
                    while (readed > 0)
                    {
                        var action = offset == 0 ? "BeginSaveChunked" : "Append";
                        await FileRequest(service, action, objectType, sessionId, projectId, batchId, 0, documentId, 0, filename,
                            offset, new ByteArrayContent(buffer, 0, readed));
                        offset += readed;
                        readed = fs.Read(buffer, 0, buffer.Length);
                        checksum = crc.Next(buffer, 0, readed);
                    }
                    var resp = await FileRequest(service, "Commit", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename);
                    if (resp.StatusCode != HttpStatusCode.OK) throw new Exception("Server error");
                }
                var response = await FileRequest(service, "Checksum", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename);
                var text = await response.Content.ReadAsStringAsync();
                if (uint.Parse(text, NumberStyles.HexNumber) != checksum)
                {
                    throw new Exception("An error occurred when uploading the file");
                }
            }
        }

        /// <summary>
        /// Sending request to FlexiCapture file service
        /// </summary>
        /// <remarks>Data encoded as x-www-form-urlencoded, saving traffic</remarks>
        /// <param name="service">Service</param>
        /// <param name="action">Requested action</param>
        /// <param name="objectType">Type of object to be processed</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="projectId">Project ID</param>
        /// <param name="batchId">Batch ID</param>
        /// <param name="parentId">Object parent ID</param>
        /// <param name="objectId">Object ID</param>
        /// <param name="version">Object version</param>
        /// <param name="streamName">Stream name</param>
        /// <returns>HTTP response from server</returns>
        private static async Task<HttpResponseMessage> FileRequest(FlexiCapture.FlexiCaptureWebServiceSoapClient service, string action,
            int objectType, int sessionId, int projectId, int batchId, int parentId, int objectId, int version, string streamName)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(service.ClientCredentials.UserName.UserName))
                creds = new NetworkCredential(service.ClientCredentials.UserName.UserName, service.ClientCredentials.UserName.Password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var uri = service.Endpoint.Address.Uri;
                var content = new FormUrlEncodedContent(new Dictionary<string, string>
                {
                    {"Action", action},
                    {"objectType", objectType.ToString()},
                    {"sessionId", sessionId.ToString()},
                    {"projectId", projectId.ToString()},
                    {"batchId", batchId.ToString()},
                    {"parentId", parentId.ToString()},
                    {"objectId", objectId.ToString()},
                    {"version", version.ToString()},
                    {"streamName", streamName},
                });
                return await client.PostAsync(uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1"), content);
            }
        }

        /// <summary>
        /// Sending request to FlexiCapture file service
        /// </summary>
        /// <remarks>Data encoded as multipart/form-data, which enables uploading files to server</remarks>
        /// <param name="service">Service</param>
        /// <param name="action">Requested action</param>
        /// <param name="objectType">Type of object to be processed</param>
        /// <param name="sessionId">Session ID</param>
        /// <param name="projectId">Project ID</param>
        /// <param name="batchId">Batch ID</param>
        /// <param name="parentId">Object parent ID</param>
        /// <param name="objectId">Object ID</param>
        /// <param name="version">Object version</param>
        /// <param name="streamName">Stream name</param>
        /// <param name="offset">File offset</param>
        /// <param name="file">File to upload</param>
        /// <returns>HTTP response from server</returns>
        private static async Task<HttpResponseMessage> FileRequest(FlexiCapture.FlexiCaptureWebServiceSoapClient service, string action,
            int objectType, int sessionId, int projectId, int batchId, int parentId, int objectId, int version, string streamName, int offset, HttpContent file)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(service.ClientCredentials.UserName.UserName))
                creds = new NetworkCredential(service.ClientCredentials.UserName.UserName, service.ClientCredentials.UserName.Password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var uri = service.Endpoint.Address.Uri;
                var content = new MultipartFormDataContent
                {
                    {new StringContent(action), "Action"},
                    {new StringContent(objectType.ToString()), "objectType"},
                    {new StringContent(sessionId.ToString()), "sessionId"},
                    {new StringContent(projectId.ToString()), "projectId"},
                    {new StringContent(batchId.ToString()), "batchId"},
                    {new StringContent(parentId.ToString()), "parentId"},
                    {new StringContent(objectId.ToString()), "objectId"},
                    {new StringContent(version.ToString()), "version"},
                    {new StringContent(Convert.ToBase64String(Encoding.Unicode.GetBytes(streamName))), "streamName"},
                };
                if (offset > 0)
                {
                    content.Add(new StringContent(offset.ToString()), "offset");
                }
                content.Add(file, "blob", "data.txt");
                return await client.PostAsync(uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1"), content);
            }
        }

    }
}

